"""Fit of the sigma timeseries with an exponential function.

This cas is to test the goodness of fit with a function that has no
lag time, i.e. the exponential function instead of a sigmoidal.
"""
import csv
from datetime import date

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends.backend_pdf import PdfPages
from scipy.optimize import curve_fit

from utils.misc import block_avg, idx_plateau
from utils.paths import ANALYSIS_PATH, CM_PATH


def exp_fun(x, A, B, k):
    return A - B * np.exp(-k * x)


# ================ Parameters ================= #
prot_name = "1ucs"
dates = ["29Oct22_CV", "2Nov22_CV"]
# prot_name = "1srl"
# dates = ["31Oct22", "7Nov22"]
b_size = 100
thr = 0.3
len_plateau = 10000 // b_size
max_var = 0.1
offset = 100 * len_plateau * b_size  # forcing fit on the whole curve

# ================ Program ================= #
datasets_used = ""
for d in dates:
    datasets_used += f"_{d}"

# load sigma timeseries
OUT_PATH = ANALYSIS_PATH / prot_name
CAVG_PATH = OUT_PATH / f"contact_avg{datasets_used}.npy"
contact_avg = np.load(CAVG_PATH)
GE_native = np.load(CM_PATH / f"GE_{prot_name}_native.npy", allow_pickle=True)

# Fit
parameters = []
para_std = []
covariance_a_b = []
avg_timeseries = []
var_timeseries = []
end_point = []

for i, lt in enumerate(GE_native):
    loop = lt[0]  # pair of residues involved in the contact

    timeseries = contact_avg[:, loop[0], loop[1]]
    t_block, t_var = block_avg(timeseries, b_size)  # reduce fluctuations
    avg_timeseries.append(t_block)
    var_timeseries.append(t_var)
    iend = idx_plateau(t_block, thr, len_plateau, max_var)

    # no plateau found; logistic's parameters are set to 0 and corresponding std to 1
    if iend is None:
        print(f"No activation plateau detected in contact {i}")
        parameters.append(np.zeros(3))
        para_std.append(np.ones(3))
        continue

    # convert index to the non-coarse-grained index and add the offset
    iend = iend * b_size + offset

    # avoid too large index
    if iend > timeseries.size - 1:
        iend = timeseries.size - 1

    xdata = np.arange(0, iend)
    ydata = contact_avg[:iend, loop[0], loop[1]]

    try:
        # constrains for the fit:
        # L in [0,1]
        # k > 0, hence no decreasing logistic
        # NOTE: the covariance matrix is computed as the inverse of the hessian matrix
        para, pcov = curve_fit(
            exp_fun,
            xdata,
            ydata,
            p0=[1, 1, 1e-3],  # initial partameters' values
            bounds=([0, 0, 0], [np.inf, np.inf, np.inf]),  # bounds
        )
    except RuntimeError as e:
        message = str(e)
        if (
            message
            == "Optimal parameters not found: Number of calls to function has reached maxfev = 800."
            or "Optimal parameters not found: The maximum number of function evaluations is exceeded."
        ):
            print(f"Contact {i} has no optimal fit to logistic")
            parameters.append(np.zeros(3))
            para_std.append(np.ones(3))
            continue
        else:
            # to avoid unexpected errors
            raise RuntimeError(message)

    perr = np.sqrt(np.diag(pcov))  # std

    covariance_a_b.append(pcov[0, 1])  # covariance between A and B
    parameters.append(para)
    para_std.append(perr)

# Save parameters in cvs file
out_name = str(OUT_PATH / f"parameters_exp_fit{datasets_used}.csv")
with open(out_name, "w") as f:
    writer = csv.writer(f)
    writer.writerow(
        [f"# Created by fit_sigma_curves_with_exp.py on date {date.today()}"]
    )
    writer.writerow(
        [
            "# contact_number",
            "contact_aa_pair",
            "A",
            "A_std",
            "B",
            "B_std",
            "k",
            "k_std",
            "cov(A,B)",
        ]
    )
    for i, lt in enumerate(GE_native):
        loop = lt[0]
        writer.writerow(
            [
                i,
                loop,
                parameters[i][0],
                para_std[i][0],
                parameters[i][1],
                para_std[i][1],
                parameters[i][2],
                para_std[i][2],
                covariance_a_b[i],
            ]
        )

# plot
total_contacts = len(GE_native)
out_name = str(OUT_PATH / "img" / f"contact_prob_fit_exp{datasets_used}.pdf")

matplotlib.rcParams["font.size"] = 20
matplotlib.rcParams["text.usetex"] = True
matplotlib.rcParams["font.family"] = "modern computer"

with PdfPages(out_name) as pdf:
    for g, lt in enumerate(GE_native):
        fig, ax = plt.subplots(1, figsize=(9, 7))
        timeseries = contact_avg[:, GE_native[g][0][0], GE_native[g][0][1]]
        lensim = len(timeseries)
        ax.set_ylim(0, 1)
        aa0 = str(lt[0][0] + 1)
        aa1 = str(lt[0][1] + 1)
        ax.set_ylabel(
            r"$\langle \sigma_{" + aa0 + ", " + aa1 + r"} \rangle (t)$", labelpad=15
        )
        ax.set_xlabel("Time [MD step]", labelpad=15)
        ax.grid(alpha=0.4)

        ax.plot(timeseries, alpha=0.2)

        xavg = np.linspace(0, lensim, lensim // b_size)
        ax.plot(xavg, avg_timeseries[g], "o", c="black", alpha=0.2)

        y = exp_fun(xavg, *parameters[g])
        ax.plot(xavg, y, c="red")

        pdf.savefig(bbox_inches="tight")
        plt.close()
