"""
Plot for the probability of contact formation with fit
for the final paper
"""
from collections import namedtuple

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import pandas as pd

from utils.misc import block_avg
from utils.paths import CM_PATH, ANALYSIS_PATH


######################## functions
def exp_fun(x, A, B, k):
    return A - B * np.exp(-k * x)


######################## main
matplotlib.rcParams["font.size"] = 24
matplotlib.rcParams["text.usetex"] = True
matplotlib.rcParams["font.family"] = "modern computer"

protein = ["1ucs", "1ucs"]
datasets = ["29Oct22_CV_2Nov22_CV", "29Oct22_CV_2Nov22_CV"]

# Contact map
cm_native = [
    np.load(CM_PATH / f"{protein[0]}_cm.npy"),
    np.load(CM_PATH / f"{protein[1]}_cm.npy"),
]

# Number of native contacts
total_contacts = [
    (np.triu(cm_native[0]) != 0).sum(),
    (np.triu(cm_native[1]) != 0).sum(),
]

# load GE for the native contacts, together with all the
# residue pairs involved in a loop (hence in a contact)
GE_native = [
    np.load(CM_PATH / f"GE_{protein[0]}_native.npy", allow_pickle=True),
    np.load(CM_PATH / f"GE_{protein[1]}_native.npy", allow_pickle=True),
]

contact_avg = [
    np.load(ANALYSIS_PATH / protein[0] / f"contact_avg_{datasets[0]}.npy"),
    np.load(ANALYSIS_PATH / protein[1] / f"contact_avg_{datasets[1]}.npy"),
]
parameters = [
    pd.read_csv(
        ANALYSIS_PATH / protein[0] / f"parameters_exp_fit_{datasets[0]}.csv",
        header=1,
        sep=",",
    ),
    pd.read_csv(
        ANALYSIS_PATH / protein[1] / f"parameters_exp_fit_{datasets[1]}.csv",
        header=1,
        sep=",",
    ),
]

idx = [38, 105]

# Plot
fig, axs = plt.subplots(2, figsize=(11, 10), gridspec_kw={"hspace": 0.25})

for ax in axs.flatten():
    ax.set_ylim(0, 1)
    ax.grid(alpha=0.4)
    ax.spines.top.set_visible(False)
    ax.spines.right.set_visible(False)
    ax.ticklabel_format(axis="x", scilimits=(0, 4))
    ax.yaxis.set_minor_locator(plt.MultipleLocator(0.1))
    ax.yaxis.set_major_locator(plt.MultipleLocator(0.2))
    ax.xaxis.set_minor_locator(plt.MultipleLocator(0.1e4))
    ax.xaxis.set_major_locator(plt.MultipleLocator(0.5e4))

# 1ucs
res1 = GE_native[0][idx[0]][0][0]
res2 = GE_native[0][idx[0]][0][1]
print(res1 + 1, res2 + 1)

timeseries = contact_avg[0][:, res1, res2]
t_block, _ = block_avg(timeseries, 100)
lensim = len(timeseries)

axs[0].plot(timeseries, alpha=0.5, c="#457b9d")
x_avg = np.linspace(0, lensim, lensim // 100)
axs[0].plot(x_avg, t_block, "o", c="#1d3557", alpha=0.3)
logis = exp_fun(
    x_avg,
    parameters[0]["A"][idx[0]],
    parameters[0]["B"][idx[0]],
    parameters[0]["k"][idx[0]],
)
axs[0].plot(x_avg, logis, c="#e63946", linewidth=2)

axs[0].set_xlim(-0.1e4, 2.60e4)
axs[0].set_ylabel(r"$\langle \sigma_{6, 25} \rangle (t)$", labelpad=15)

# second fit
res1 = GE_native[1][idx[1]][0][0]
res2 = GE_native[1][idx[1]][0][1]
print(res1 + 1, res2 + 1)

timeseries = contact_avg[1][:, res1, res2]
t_block, _ = block_avg(timeseries, 100)
lensim = len(timeseries)

(l1,) = axs[1].plot(timeseries, alpha=0.5, c="#457b9d")
x_avg = np.linspace(0, lensim, lensim // 100)
(l2,) = axs[1].plot(x_avg, t_block, "o", c="#1d3557", alpha=0.3)
logis = exp_fun(
    x_avg,
    parameters[1]["A"][idx[1]],
    parameters[1]["B"][idx[1]],
    parameters[1]["k"][idx[1]],
)
(l3,) = axs[1].plot(x_avg, logis, c="#e63946", linewidth=2)

axs[1].set_xlim(-0.1e4, 2.60e4)
axs[1].set_ylabel(r"$\langle \sigma_{23, 37} \rangle (t)$", labelpad=15)
axs[1].set_xlabel("Time [MD step]", labelpad=15)

# legend
axs[1].legend(
    handles=[l1, l2, l3],
    labels=[
        "Timeseries",
        "Block average",
        "Fit $A - B \\exp \\left( - k t \\right)$",
    ],
    loc="upper center",
    bbox_to_anchor=(0.45, 2.55),
    ncol=3,
    fancybox=True,
    shadow=True,
    fontsize=20,
)

fig.savefig(
    ANALYSIS_PATH
    / "1ucs"
    / "img"
    / f"contact_probability_exp_fit_1ucs_contacts_{idx[0]}_1ucs_contact_{idx[1]}.pdf",
    bbox_inches="tight",
)
