import os
from typing import List

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import visualization.viztool as plotf
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy.stats import pearsonr, spearmanr
from sklearn.linear_model import LinearRegression
from utils.paths import ANALYSIS_PATH, CM_PATH, DATA_PATH

# graphical settings
matplotlib.rcParams["font.size"] = 24
matplotlib.rcParams["text.usetex"] = True
matplotlib.rcParams["font.family"] = "modern computer"

#################### Parameters ####################
prot_name = "1ucs"
dates = ["29Oct22_CV", "2Nov22_CV"]
dataset_unfolded = "30Oct22_CV"
temp_unfolded = "1_250"
####################################################

# where results are saved
OUT_PATH = ANALYSIS_PATH / prot_name
# tag
datasets_used = ""
for d in dates:
    datasets_used += f"_{d}"

# Contact map
cm_native = np.load(CM_PATH / f"{prot_name}_cm.npy")
cm_native = np.triu(cm_native)
n_res = cm_native.shape[0]

# load GE for the native contacts, together with all the
# residue pairs involved in a loop (hence in a contact)
GE_native = np.load(CM_PATH / f"GE_{prot_name}_native.npy", allow_pickle=True)

# used to have a reference for the len of the simulations
Q = np.load(OUT_PATH / f"Q{datasets_used}.npy")
id_map = np.loadtxt(
    OUT_PATH / f"ID_mapping{datasets_used}.csv", delimiter=",", dtype=int
)

# Load parameters from fit
parameters = pd.read_csv(
    OUT_PATH / f"parameters_exp_fit{datasets_used}.csv", header=1, sep=","
)

### Pre-processing
### Definition for 1ucs plots
# See intermediate/IEm_IT_contact_selection.py for the details
# z-score ie > 1 and z-score it < 1
thr_contacts = [
    (0, 26),
    (1, 25),
    (1, 29),
    (1, 34),
    (2, 29),
    (2, 31),
    (2, 34),
    (3, 24),
    (3, 26),
    (3, 29),
    (3, 31),
    (4, 32),
    (5, 24),
    (5, 26),
]
# See intermediate/IEm_first_entangled_contacts_selection.py for the details
# Z-score > 0
early_entangled = [(21, 39), (22, 36), (23, 36)]

para_len = parameters.shape[0]
assert para_len == len(GE_native)

### Compute probability
# NOTE: in this document, tstar is equal to the B factor in the exponential fit
# A-B*exp(-k*t)
df_contact_former_U = pd.read_csv(
    OUT_PATH / f"formed_contacts_U_{dataset_unfolded}_{temp_unfolded}.csv",
    header=0,
    sep=",",
)

# save to managable tstar dict (here we use the already developed code)
tstar_all = {
    "value": np.zeros(para_len),
    "ci": np.zeros((2, para_len)),
    "aa_pair": [],
}

for idx, aa_pair in enumerate(parameters["contact_aa_pair"]):
    aa_pair_spl = aa_pair.split(",")
    aa0 = int(aa_pair_spl[0].strip("("))
    aa1 = int(aa_pair_spl[1].strip(")").strip(" "))

    probability = (
        df_contact_former_U.loc[
            df_contact_former_U["aa_pair"] == aa_pair, "n_formed"
        ].values[0]
        / df_contact_former_U.loc[
            df_contact_former_U["aa_pair"] == aa_pair, "n_total"
        ].values[0]
    )
    tstar_all["value"][idx] = probability
    # This error is the STD of the stochastic p variable, assumed binomially distributed
    # I.e. (p(1-p)/N)^1/2
    err = (
        probability
        * (1 - probability)
        / df_contact_former_U.loc[
            df_contact_former_U["aa_pair"] == aa_pair, "n_total"
        ].values[0]
    ) ** 0.5

    err_low = err
    err_high = err
    # Prevent errors to be less than 0 or greater than 1
    if probability - err_low < 0:
        err_low = probability
    if probability + err_high > 1:
        err_high = 1 - probability

    tstar_all["ci"][:, idx] = np.array([err_low, err_high])
    tstar_all["aa_pair"].append((aa0, aa1))

### k and GE
GE_all = np.zeros(para_len)
k_all = {"value": np.zeros(para_len), "std": np.zeros(para_len)}
thr_contacts_identifier = []
early_entangled_contacts_identifier = []

for idx, (contact_pair, k, a, dk, da, ge_l, aa_pair_check_p) in enumerate(
    zip(
        parameters["contact_aa_pair"],
        parameters["k"],
        parameters["A"],
        parameters["k_std"],
        parameters["A_std"],
        GE_native,
        tstar_all["aa_pair"],
    )
):
    contact_pair = contact_pair.split(",")
    aa0 = int(contact_pair[0].strip("("))
    aa1 = int(contact_pair[1].strip(")").strip(" "))
    assert aa0 == ge_l[0][0] and aa1 == ge_l[0][1]
    assert aa0 == aa_pair_check_p[0] and aa1 == aa_pair_check_p[1]

    GE_all[idx] = abs(ge_l[2])
    # case in which no plateau or fit are found
    # tstar is equal to the len of the simulation
    if k == 0:
        k_all["value"][idx] = 0
        k_all["std"][idx] = 0
    else:
        k_all["value"][idx] = k
        k_all["std"][idx] = dk

    # identify which contacts are part of the "thread and loop" region
    if ge_l[0] in thr_contacts:
        thr_contacts_identifier.append(1)
    else:
        thr_contacts_identifier.append(0)
    if ge_l[0] in early_entangled:
        early_entangled_contacts_identifier.append(1)
    else:
        early_entangled_contacts_identifier.append(0)
assert len(GE_all) == len(tstar_all["value"])
assert len(GE_all) == len(k_all["value"])
assert tstar_all["ci"].shape[1] == len(k_all["std"])

### Correlation between $|G'|$ and $t^*$
IMG_PATH = OUT_PATH / "img/"
TXT_PATH = OUT_PATH / "txt/"
if not os.path.isdir(IMG_PATH):
    os.mkdir(IMG_PATH)
if not os.path.isdir(TXT_PATH):
    os.mkdir(TXT_PATH)

fig, ax = plt.subplots(1, figsize=(9, 7))
plotf.err_scatter(
    ax, GE_all, tstar_all["value"], tstar_all["ci"], "$|G'|$", "$p_{init}$"
)
fig.savefig(
    IMG_PATH / f"GE_p_U_prefiltering_exp{datasets_used}.pdf", bbox_inches="tight"
)


# save info about correlation and removed points
out_corr = open(TXT_PATH / f"GE_p_U_k_correlation_exp{datasets_used}.txt", "w")

# filtering
l_thr = -10000  # NOTE: in this version, this is useless

noplateau_nooptimal_filtered = []
l_thr_filtered = []
idx = 0
for ts, ge in zip(tstar_all["value"], GE_all):
    if ts == Q[0].size:
        noplateau_nooptimal_filtered.append(idx)
    elif ts < l_thr:
        l_thr_filtered.append(idx)
    idx += 1

all_filtered = noplateau_nooptimal_filtered + l_thr_filtered
mask_tstar = np.ones(GE_all.size, dtype=bool)
mask_tstar[all_filtered] = False

# removed
s = (
    f"Contacts removed: {len(all_filtered)}\n"
    + f"No plateau, no optimal fit: {noplateau_nooptimal_filtered}\n"
    + f"Lower than l_thr: {l_thr_filtered}\n"
)
print(s)
out_corr.write(s)

# compute Pearson's r
r, p_r = pearsonr(GE_all, tstar_all["value"])
# compute Spearman's r
r_s, p_s = spearmanr(GE_all, tstar_all["value"])
s = (
    f"\nConsidering ALL contacts\n"
    + "---------------------------------------------\n"
    + f"Pearson's r = {r:.3f}; p-value = {p_r:.3E}\n"
    + f"Spearman's r = {r_s:.3f}; p-value = {p_s:.3E}\n"
)
print(s)
out_corr.write(s)

r, p_r = pearsonr(GE_all[mask_tstar], tstar_all["value"][mask_tstar])
r_s, p_s = spearmanr(GE_all[mask_tstar], tstar_all["value"][mask_tstar])
s = (
    f"\nContacts with p_U lower that {l_thr} and without fit REMOVED\n"
    + "---------------------------------------------\n"
    + f"Pearson's r = {r:.3f}; p-value = {p_r:.3E}\n"
    + f"Spearman's r = {r_s:.3f}; p-value = {p_s:.3E}\n"
)
print(s)
out_corr.write(s)

fig, ax = plt.subplots(1, figsize=(9, 7))
plotf.err_scatter(
    ax,
    GE_all[mask_tstar],
    tstar_all["value"][mask_tstar],
    tstar_all["ci"] * mask_tstar[np.newaxis, :],
    "$|G'|$",
    r"$p_\mathrm{U}$",
)
# linear regression
reg = LinearRegression().fit(
    np.array(GE_all[mask_tstar]).reshape(-1, 1),
    np.array(tstar_all["value"][mask_tstar]).reshape(-1, 1),
)
print(f"Linear regression Slope = {reg.coef_[0][0]:.2E}")

regressionplot = True
if regressionplot == True:
    # drawing linear regression
    x = np.linspace(min(GE_all[mask_tstar]), max(GE_all[mask_tstar]), 100)
    y = x * reg.coef_[0][0] + reg.intercept_[0]
    ax.plot(x, y, "--", color="#ff5100")  # complementary color to deepskyblue

# adjust plot
plotf.textBox(ax, x=0.8, y=0.15, s=prot_name)
plotf.set_tickers(ax)
ax.ticklabel_format(style="sci", useMathText=True, scilimits=(0, 0))

fig.tight_layout()
fig.savefig(IMG_PATH / f"GE_p_U_exp{datasets_used}.pdf")

### Correlation between $k$ and $|G'|$

# filtering
k_filtered = []
u_thr = 100
for i, (k, dk, ge) in enumerate(zip(k_all["value"], k_all["std"], GE_all)):
    if k == 0 or k > u_thr:
        k_filtered.append(i)
print(f"k values filtered: {k_filtered}")
mask_k = np.ones(GE_all.size, dtype=bool)
mask_k[k_filtered] = False

## $t^*$ and $k$ correlation
def err_scatter(
    ax: plt.Axes,
    x: List[float],
    y: List[float],
    xerr: List[float],
    yerr: List[float],
    xlabel: str,
    ylabel: str,
) -> None:
    """Scatter plot with error bars on the y axis"""
    # sort arrays
    # x_sorted, y_sorted, x_err_sorted, y_err_sorted = zip(*sorted(zip(x, y, xerr, yerr)))

    # axs[0].scatter(qx, gey, c=time_endstep, marker='o', linewidths=7, cmap=cmap, s=10)
    ax.errorbar(
        x,
        y,
        xerr=xerr,
        yerr=yerr,
        linestyle="none",
        fmt=".",
        markersize=7,
        capsize=3,
        ecolor="black",
    )

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)


def scatter_color(
    ax: plt.Axes,
    x,
    y,
    color,
    cmap,
    xlabel: str,
    ylabel: str,
    bar_label: str,
) -> None:
    """Scatter plot with error bars on the y axis"""
    assert len(x) == len(y)
    assert len(x) == len(color)

    s = ax.scatter(x, y, c=color, cmap=cmap, marker="o", linewidths=7, s=10)

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.75)
    c = fig.colorbar(s, ax=ax, cax=cax)
    c.set_label(bar_label, rotation=0, labelpad=25)

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    return s


# defining a mask for both tstar and k filtered
mask_ts_k = mask_tstar & mask_k

# remove from thr_contacts_identifier those that
# are removed from the tstart and k lists
# NB: this is true is the contact belongs to the group
# and has a tstart and k parameter
thr_contacts_identifier_mask = mask_ts_k & np.array(thr_contacts_identifier).astype(
    bool
)
early_entangled_contacts_identifier_mask = mask_ts_k & np.array(
    early_entangled_contacts_identifier
).astype(bool)

# plot
fig, ax = plt.subplots(1, figsize=(9, 7))
cmap = sns.color_palette("mako_r", as_cmap=True)

sc_plot = scatter_color(
    ax,
    k_all["value"][mask_ts_k],
    tstar_all["value"][mask_ts_k],
    GE_all[mask_ts_k],
    cmap,
    "$k$ [1/MD step]",
    r"$p_\mathrm{U}$",
    "$|G'|$",
)
# Add error in the k direction only for points with std bigger than the marker size
marker_size = sc_plot.get_sizes()[0] / 2 - 1  # -1 corrects some error not displaying
marker_size_xscale = marker_size / fig.dpi / fig.get_size_inches()[0] * ax.get_xlim()[1]

std_mask = []
for std in k_all["std"]:
    if std > marker_size_xscale:
        std_mask.append(True)
    else:
        std_mask.append(False)

std_mask = np.array(std_mask) & mask_ts_k
print(std_mask.sum())
ax.errorbar(
    k_all["value"][std_mask],
    tstar_all["value"][std_mask],
    xerr=k_all["std"][std_mask],
    linestyle="none",
    fmt="",
    c="black",
    markersize=2,
    capsize=3,
    ecolor="black",
    alpha=0.8,
    zorder=0
)
# thread contacts
ax.scatter(
    k_all["value"][thr_contacts_identifier_mask],
    tstar_all["value"][thr_contacts_identifier_mask],
    s=75,
    facecolors="none",
    edgecolors="r",
)
# early entangled contacts
ax.scatter(
    k_all["value"][early_entangled_contacts_identifier_mask],
    tstar_all["value"][early_entangled_contacts_identifier_mask],
    s=75,
    facecolors="none",
    edgecolors="fuchsia",
)
# ticks
ax.ticklabel_format(style="sci", axis="x", useMathText=True, scilimits=(0, 0))
# other options
ax.grid(alpha=0.5)

plotf.textBox(ax, x=0.8, y=0.15, s="RD1")

print("Ax y lim of main plot: ", ax.get_ylim())
fig.savefig(
    IMG_PATH / f"p_U_k_exp{datasets_used}_highlight_thread_contacts_TSE_w_GE.pdf",
    bbox_inches="tight",
)

# compute Pearson's r
r, p_r = pearsonr(tstar_all["value"][mask_ts_k], k_all["value"][mask_ts_k])
# compute Spearman's r
r_s, p_s = spearmanr(tstar_all["value"][mask_ts_k], k_all["value"][mask_ts_k])

s = (
    "\nk p_U correlation\n"
    + "---------------------------------------------\n"
    + f"Pearson's r = {r:.3f}; p-value = {p_r:.3E}\n"
    + f"Spearman's r = {r_s:.3f}; p-value = {p_s:.3E}\n"
)
print(s)
out_corr.write(s)
out_corr.close()


## Last plot with errors: review
# plot
fig, axs = plt.subplots(1, 2, figsize=(18, 7))
cmap = sns.color_palette("mako_r", as_cmap=True)

def scatter_color(
    ax: plt.Axes,
    x,
    y,
    color,
    cmap,
    xlabel: str,
    ylabel: str,
    bar_label: str,
) -> None:
    """Scatter plot with error bars on the y axis"""
    assert len(x) == len(y)
    assert len(x) == len(color)

    s = ax.scatter(x, y, c=color, cmap=cmap, marker="o", linewidths=7, s=5, alpha=0.5)

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.2)
    c = fig.colorbar(s, ax=ax, cax=cax)
    c.set_label(bar_label, rotation=0, labelpad=25)

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

# error on y
scatter_color(
    axs[0],
    k_all["value"][mask_ts_k],
    tstar_all["value"][mask_ts_k],
    GE_all[mask_ts_k],
    cmap,
    "$k$ [1/MD step]",
    r"$p_\mathrm{U}$",
    "",
)
axs[0].errorbar(
    k_all["value"][mask_ts_k],
    tstar_all["value"][mask_ts_k],
    yerr=tstar_all["ci"] * mask_ts_k[np.newaxis, :],
    linestyle="none",
    fmt=".",
    c="black",
    markersize=2,
    capsize=3,
    ecolor="black",
    alpha=0.5
)
# thread contacts
axs[0].scatter(
    k_all["value"][thr_contacts_identifier_mask],
    tstar_all["value"][thr_contacts_identifier_mask],
    s=75,
    facecolors="none",
    edgecolors="r",
    alpha=0.5,
)

# early entangled contacts
axs[0].scatter(
    k_all["value"][early_entangled_contacts_identifier_mask],
    tstar_all["value"][early_entangled_contacts_identifier_mask],
    s=75,
    facecolors="none",
    edgecolors="fuchsia",
    alpha=0.5,
)

# ticks
axs[0].ticklabel_format(style="sci", axis="both", useMathText=True, scilimits=(0, 0))
# other options
axs[0].grid(alpha=0.5)

plotf.textBox(
    axs[0],
    x=0.45,
    y=0.925,
    fontsize=17,
    s=r"RD1: $p_\mathrm{U}$ std",
)

# error on x
scatter_color(
    axs[1],
    k_all["value"][mask_ts_k],
    tstar_all["value"][mask_ts_k],
    GE_all[mask_ts_k],
    cmap,
    "$k$ [1/MD step]",
    "",
    "$|G'|$",
)
axs[1].errorbar(
    k_all["value"][mask_ts_k],
    tstar_all["value"][mask_ts_k],
    xerr=k_all["std"][mask_ts_k],
    linestyle="none",
    fmt=".",
    c="black",
    markersize=2,
    capsize=3,
    ecolor="black"
)
# thread contacts
axs[1].scatter(
    k_all["value"][thr_contacts_identifier_mask],
    tstar_all["value"][thr_contacts_identifier_mask],
    s=75,
    facecolors="none",
    edgecolors="r",
    alpha=0.5,
)
# early entangled contacts
axs[1].scatter(
    k_all["value"][early_entangled_contacts_identifier_mask],
    tstar_all["value"][early_entangled_contacts_identifier_mask],
    s=75,
    facecolors="none",
    edgecolors="fuchsia",
    alpha=0.5,
)

# ticks
axs[1].ticklabel_format(style="sci", axis="both", useMathText=True, scilimits=(0, 0))
# other options
axs[1].grid(alpha=0.5)

plotf.textBox(axs[1], x=0.75, y=0.925, fontsize=17, s="RD1: $k$ std")

# adjustments
axs[1].yaxis.set_ticklabels([])

fig.savefig(
    IMG_PATH
    / f"p_U_k_exp{datasets_used}_highlight_thread_contacts_TSE_w_GE_errors.pdf",
    bbox_inches="tight",
)
