"""Computes contact formation probabilities in the unfolded ensemble"""
import MDAnalysis as mda
import numpy as np
import pandas as pd
from utils.contacts import formed_contacts
from utils.paths import ANALYSIS_PATH, CM_PATH, DATA_PATH


def formed_contacts_timeseries(
    cm_native: np.ndarray,
    u: mda.Universe,
    gamma: float,
    begin: int,
    end: int,
    step: int,
) -> np.ndarray:
    """Returns the series of matrix representing formed contacts as a function
    of time.

    Args:
        cm_native (2D numpy array): contact map of the native state of the protein
        u (MDAnalysis Universe): Universe representing the trajectory
        gamma (float): parameter to define contact
        begin, end (int): start and end of the trajectory range to analyse
        If end == 0, the whole trajectory is considered

    Returns:
        1D Numpy array: the contact timeseries
    """

    if end > u.trajectory.n_frames or end == 0:
        end = u.trajectory.n_frames

    n_res = cm_native.shape[0]
    n_configurations = len(list(range(0, end - begin, step)))
    timeseries = np.zeros((n_configurations, n_res, n_res))

    idx = 0
    for i in range(0, end - begin, step):
        ca_pos = u.trajectory[i].positions
        timeseries[idx] = formed_contacts(cm_native, ca_pos, gamma)
        idx += 1
    assert idx == n_configurations 
    # not -1 because we update idx in the last step of the loop

    return timeseries


#################### Parameters ####################
# prot_name = "1ucs"
# dataset_unfolded = "30Oct22_CV"
# temp = "1_250"
dates = ["29Oct22_CV", "2Nov22_CV"]
prot_name = "1srl"
dataset_unfolded = "30Oct22"
temp = "1_200"
dates = ["31Oct22", "7Nov22"]

sub_sampling_step = 50  # used to avoid correlations
start = 8000  # used to remove equilibration
####################################################
print(f"Computing contact map for {prot_name}, {dataset_unfolded} dataset at {temp}")

OUT_PATH = ANALYSIS_PATH / prot_name
# tag
datasets_used = ""
for d in dates:
    datasets_used += f"_{d}"
parameters = pd.read_csv(
    OUT_PATH / f"parameters_exp_fit{datasets_used}.csv", header=1, sep=","
)
df = pd.DataFrame(columns=["aa_pair", "n_formed", "n_total"])
for aa_pair in parameters["contact_aa_pair"]:
    df.loc[len(df), :] = [aa_pair, 0, 0]

# Load trajectory
PSF = (
    DATA_PATH
    / f"{dataset_unfolded}"
    / f"Data_{prot_name}_{temp}"
    / f"{prot_name}_init.data"
)
eqDCD = (
    DATA_PATH / f"{dataset_unfolded}" / f"Data_{prot_name}_{temp}" / f"{prot_name}.dcd"
)
universe = mda.Universe(str(PSF), str(eqDCD))

# used to have a reference for the len of the simulations
cm_native = np.load(CM_PATH / f"{prot_name}_cm.npy")
cm_native = np.triu(cm_native)

print(
    f"Start contact computation for {len(list(range(0, universe.trajectory.n_frames - start, sub_sampling_step)))} configurations"
)
contact_map_timeseries = formed_contacts_timeseries(
    cm_native, universe, 1.2, start, 0, sub_sampling_step
)

# Count contacts and asve dataframe
n_configurations = contact_map_timeseries.shape[0]
contact_map_formed_contacts = contact_map_timeseries.sum(axis=0)
for aa_pair in parameters["contact_aa_pair"]:
    aa_pair_spl = aa_pair.split(",")
    aa0 = int(aa_pair_spl[0].strip("("))
    aa1 = int(aa_pair_spl[1].strip(")").strip(" "))
    if contact_map_formed_contacts[aa0, aa1] != 0:
        df.loc[
            df["aa_pair"] == aa_pair, "n_formed"
        ] = contact_map_formed_contacts[aa0, aa1]
    df.loc[
            df["aa_pair"] == aa_pair, "n_total"
    ] = n_configurations

df.to_csv(OUT_PATH / f"formed_contacts_U_{dataset_unfolded}_{temp}.csv", index=False)