"""
Classification of different pathways for 1ucs

This classification is based on the Q, <G'> landscape
"""
import csv

import numpy as np


def convolution(timeseries, window=5):
    return np.convolve(timeseries, np.ones(window) / window, mode="valid")


def timeseries_mask(r: int, t: np.ndarray) -> np.ndarray:
    """Return a mask applicable to timeseries t;
    The first and the last point in the timeseries are selected

    Please note that if you want to use mask for fancy indexing
    you need to use np.nonzero(mask) instead"""

    assert t.ndim == 1

    mask = np.zeros(t.size, dtype=int)
    block = mask.size // r
    for i in range(r):
        mask[i * block] = 1
    mask[-1] = 1  # include last frame

    return mask


# definition ensembles
# Q, GE
IT = ((0.6, 0.75), (-0.2, 0.2))
N = ((0.7, 1.0), (-1, -0.5))
U = ((0, 0.5), (-0.25, 0.25))


# load data
Q = np.load("/mnt/GO_ANALYSIS/1ucs/Q_29Oct22_CV_2Nov22_CV.npy", allow_pickle=True)
GE = np.load(
    "/mnt/GO_ANALYSIS/1ucs/wGEa_threshold_0.50_exponent_3.00_r1500_29Oct22_CV_2Nov22_CV.npy",
    allow_pickle=True,
)
ids = np.loadtxt(
    "/mnt/GO_ANALYSIS/1ucs/ID_mapping_29Oct22_CV_2Nov22_CV.csv",
    delimiter=",",
    skiprows=3,
    dtype=int,
)

r = GE.shape[1] - 1
mask = timeseries_mask(r, Q[0])
nonzero_mask = np.flatnonzero(mask)

win = 20

n_fast = 0
n_thr = 0
n_bt = 0
n_trapped = 0

classifications = []

for q, ge, map in zip(Q[:], GE, ids):
    print(f"Trajectory {map}")
    class_temp = [map]

    q_under = np.array(q)[mask.astype(bool)]
    assert q_under.size == ge.size

    # convolution
    q_conv = convolution(q_under, win)
    ge_conv = convolution(ge, win)

    # classify trajectory
    # U -> -1
    # IT -> 1
    # N -> 2
    classification = np.zeros(q_conv.size, dtype=int)
    classification[
        np.where(
            (q_conv > U[0][0])
            & (q_conv < U[0][1])
            & (ge_conv > U[1][0])
            & (ge_conv < U[1][1])
        )
    ] = -1
    classification[
        np.where(
            (q_conv > IT[0][0])
            & (q_conv < IT[0][1])
            & (ge_conv > IT[1][0])
            & (ge_conv < IT[1][1])
        )
    ] = 1
    classification[
        np.where(
            (q_conv > N[0][0])
            & (q_conv < N[0][1])
            & (ge_conv > N[1][0])
            & (ge_conv < N[1][1])
        )
    ] = 2

    # fast folding case
    if not np.any(classification == 1) and np.any(classification == 2):
        print("Fast folding")
        n_fast += 1
        class_temp.append("fast")
    elif not np.any(classification == 2):
        # it didn't fold
        print("Trapped")
        n_trapped += 1
        class_temp.append("trapped")
    else:
        # check if it is a THR
        idx_folded = np.where(classification == 2)[0]
        if idx_folded.size > 0:
            if np.any(classification[idx_folded[0] - 25: idx_folded[0]] == 1):
                print("THR")
                n_thr += 1
                class_temp.append("thr")
            # BT case
            elif np.any(classification[idx_folded[0] - 25: idx_folded[0]] == -1):
                print("BT")
                n_bt += 1
                class_temp.append("bt")
            else:
                raise ValueError("No class found")
        else:
            raise ValueError("No class found")

    classifications.append(class_temp)
    print("============")

print(f"Fast: {n_fast}")
print(f"THR: {n_thr}")
print(f"BT: {n_bt}")
print(f"Trapped: {n_trapped}")
print("Total trajectories:", n_fast + n_thr + n_bt + n_trapped)

# save classifications
with open(
    "/mnt/GO_ANALYSIS/1ucs/classifications_traj_29Oct22_CV_2Nov22_CV.csv",
    "w",
    newline="",
) as csvfile:
    writer = csv.writer(csvfile, delimiter=",")
    writer.writerow(["# Generated by 1ucs_automatic_classification.py"])
    writer.writerow(["# index", "id_run", "classification"])
    for row in classifications:
        writer.writerow([row[0][0], row[0][1], row[1]])
