"""
Script to generate list of Gaussian Entanglements from trajectories.
For each snapshot in each trajectory, the scripts compute the contacts formed and
compute G' for each of these.
For each trajectory, it save this list related to each snapshot.

The user specifies the number of snapshots (equally spaced) to consider through
the "r" variable
"""
import os
from pathlib import Path

import numpy as np
import MDAnalysis as mda

from utils.GE import trajectory_GE_loops
from utils.misc import timeseries_mask
from utils.paths import ANALYSIS_PATH, DATA_PATH, CM_PATH

#################################
prot_name = '1ucs'

dates = [
    '29Oct22_CV',
    '2Nov22_CV'
]
gamma = 1.2
r = 1500 # number configurations analyzed for each trajectory
m0 = 10 # minimum thread len

checkpoints = True
#################################

# where to save results
OUT_PATH = ANALYSIS_PATH / prot_name

# to tag results
datasets_used = ''
for d in dates:
    datasets_used += f'_{d}'

# Directory with the trajectory
simulations = [str(DATA_PATH / date / sim)
               for date in dates
               for sim in os.listdir(DATA_PATH / date)
               if sim != 'logs']

# Contact map
cm_native = np.load(CM_PATH / f'{prot_name}_cm.npy')
cm_native = np.triu(cm_native)

# Universes for all simulations
# NOTE: this is NOT memory expensive, because MDA do not load the
# entire trajectory all at once
universes = []
for sim in simulations:
    simdir = Path(sim)
    PSF = simdir / f'{prot_name}_init.data'
    eqDCD = simdir / f'{prot_name}.dcd'
    universes.append(mda.Universe(str(PSF), str(eqDCD)))
universes = np.array(universes)

Q = np.load(OUT_PATH / f'Q{datasets_used}.npy', allow_pickle=True)
id_map = np.loadtxt(OUT_PATH / f"ID_mapping{datasets_used}.csv", delimiter=",", dtype=int)

# Checks and reloading of simulations and universes if Q and RMSD has been loaded
# from disk
assert Q.shape[0] == id_map.shape[0]
# this means that the files has been loaded
# Hence, simulations and universes need to be ordered
# in the same manner
assert len(simulations) == id_map.shape[0]
for sim, id_tuple in zip(simulations, id_map):
    id_from_sim = int(sim.split("_")[-1])
    if id_tuple[1] != id_from_sim:
        print(f"Change {id_from_sim} with {id_tuple[1]} in position {id_tuple[0]}")

        str_split = simulations[id_tuple[0]].split("_")[:-1]
        if id_tuple[1] < 500:
            str_split[-1] = str(id_tuple[1])
        # this is made to handle temperatures runs, where the id_map have an ID
        # > 500 (hopefully)
        elif id_tuple[1] >= 500:
            id_str = str(id_tuple[1])
            str_split[-1] = f"{id_str[0]}_{id_str[1:]}"
        else:
            raise RuntimeError("Control ID_map\n")
        simulations[id_tuple[0]] = "_".join(str_split)

        simdir = Path(simulations[id_tuple[0]])
        PSF = simdir / f'{prot_name}_init.data'
        eqDCD = simdir / f'{prot_name}.dcd'
        universes[id_tuple[0]] = mda.Universe(str(PSF), str(eqDCD))

# NOTE: With this choice, all samplings are fixed by the first timeseries
# Hence, for different runs' lens the "sampling rate" will be constant.
# Be aware that this behavior is under control for SHORTER runs but NOT
# for LONGER one
mask = timeseries_mask(r, Q[0])

# compute GE
GE_PATH = OUT_PATH / f'GE4all_loops_r{r}{datasets_used}.npy'

# save a set of arrays if checkpoints are selected
if checkpoints:
    GE_PATH = OUT_PATH / f"GE4all_loops_r{r}{datasets_used}" # a directory
    os.mkdir(GE_PATH)

GE_traj = []
k = 0
for u, id_tuple in zip(universes, id_map):
    print(f'Loading: {k}')
    k += 1

    GE_loops_traj = trajectory_GE_loops(u, mask, cm_native, gamma, m0)
    if checkpoints:
        np.save(GE_PATH / f'run_{id_tuple[1]}.npy', GE_loops_traj)
    else:
        GE_traj.append(GE_loops_traj)

if not checkpoints:
    np.save(GE_PATH, GE_traj)
