'''Functions to analyse topological complexes in protein configurations
through the Gaussian Entanglement as defined in baiesi2019
'''
import numpy as np
from MDAnalysis.analysis import distances

from .contacts import formed_contacts
from .GE_routine import computeGE # cython implementation

def find_loops(cm):
    '''Return list of two numbers, each defining the beginning and the
    end of a loop in a single polypeptide chain'''
    # select only upper triangular matrix
    cm_triu = np.triu(cm, 1)
    # in a single polypeptide chain, each contact creates a loop
    idx = np.where(cm_triu != 0)

    return np.hstack((idx[0][...,None], idx[1][...,None]))

def compute_R(ca_positions):
    '''Computes R vectors for GE calculation, i.e. bead positions'''
    return ((ca_positions[1:] + ca_positions[:-1])/2).astype(float)

def compute_deltaR(ca_positions):
    '''Computes DeltaR for GE calculation, i.e. bond vectors'''
    return (ca_positions[1:] - ca_positions[:-1]).astype(float)

def GE_loops(cm, ca_positions, m0):
    '''Compute G' for all loops, i.e. contacts, in a single polypeptide chain

    Args:
        cm (2D numpy array): contact map for the configuration
        ca_positions (2D numpy array): list of the 3 dimensional position for
        each C_alpha (or residue in coarse-grained fashion)
        m0 (int): minimum length for a thread

    Returns:
        list of N by 3 elements: first is a tuple characterizing the loops, 
        the second related to thread and the last one is the GE values associated
    '''

    ca_positions = np.array(ca_positions)
    len_chain = ca_positions.shape[0]
    loops = find_loops(cm) 
    # note: first index is 0
    # note: second index > first index, for construction

    R = compute_R(ca_positions)
    deltaR = compute_deltaR(ca_positions)

    GE = []
    j_list = []
    for loop in loops:

        # loop extremes
        i1 = loop[0]
        i2 = loop[1]

        GE_max = 0
        j1_max = 0
        j2_max = 0
        # before the loop
        for j1 in range(i1-m0):
 
            for increment in range(i1 - m0 - j1):
                j2 = j1 + m0 + increment
                GE_loop = computeGE(i1,i2,j1,j2, R, deltaR)
                if abs(GE_loop) > abs(GE_max):
                    GE_max = GE_loop
                    j1_max = j1
                    j2_max = j2

        # after the loop
        for j1 in range(i2 + 1, len_chain - m0): 

            for increment in range(len_chain - m0 - j1):
                j2 = j1 + m0 + increment
                GE_loop = computeGE(i1,i2,j1,j2, R, deltaR)
                if abs(GE_loop) > abs(GE_max):
                    GE_max = GE_loop
                    j1_max = j1
                    j2_max = j2


        # loop's GE is the maximum in modulus
        GE.append(GE_max)
        j_list.append((j1_max, j2_max))

    # fix output
    result = []
    for i, j, ge in zip(loops, j_list, GE):
        result.append([tuple(i), j, ge])

    return result

def GE_config(ge_loops, m0):
    '''Returns the GE of a protein and the corresponding loop-thread.
    The protein's GE is the absolute maximum among all GEs

    Args:
        ge_loops (list of N by 3 elements): output of GE_loops()
        m0 (int): minimum loop len
    
    Returns:
        [(loop), (thread), GE_protein]
    '''

    l_ge = [x for x in ge_loops if abs(x[0][1]-x[0][0]) >= m0]
    if len(l_ge) == 0:
        GE = 0
        loop = None
        thread = None
    else:
        ge = [y[2] for y in l_ge]
        ge_max = max(ge, key=abs)
        GE = ge_max
        loop = l_ge[ge.index(ge_max)][0]
        thread = l_ge[ge.index(ge_max)][1]

    return [loop, thread, GE]


def mGE_config(ge_loops, m0):
    '''Returns the GE of a protein and the corresponding loop-thread.
    The protein's GE is the average among all GEs

    Args:
        ge_loops (list of N by 3 elements): output of GE_loops()
        m0 (int): minimum loop len

    Returns:
        [(loop), (thread), GE_protein]
    '''

    l_ge = [x for x in ge_loops if abs(x[0][1]-x[0][0]) >= m0]
    if len(l_ge) == 0:
        GE = 0
        loop = None
        thread = None
    else:
        ge = [y[2] for y in l_ge]
        ge_mean = np.mean(ge)
        GE = ge_mean
        loop = None #l_ge[ge.index(ge_mean)][0]
        thread = None #l_ge[ge.index(ge_mean)][1]

    return [loop, thread, GE]


def trajectory_GE_loops(u, mask, cm_native, gamma, m0):
    '''Compute GE for each formed contact, hence loop, in selected configurations 
    of a trajectory. The mask argument is used fot select these configurations.

    Args:
        u (MDAnalysis Universe): Universe of the trajectory
        mask (list of bool): list with the configurations to analysis,
        they can be uncorrelated using unfold.subsample_unfold()
        cm_native (2D numpy array): contact map of the native state
        gamma (float): parameter to define a native contact
        m0 (int): number of residues to considere a chain segment a "thread"
    '''

    GE_traj = []
    for i in range(u.trajectory.n_frames):
        if mask[i]:
            print(f'Computing GE for config: {i}')
            ca_pos = u.trajectory[i].positions
            cm_config = formed_contacts(cm_native, ca_pos, gamma)
            GE_traj.append(GE_loops(cm_config, ca_pos, m0))

    return GE_traj

def trajectory_GE_loops_nonnative(u, mask, thr, gamma, m0):
    '''Compute GE for each formed contact, hence loop, in selected configurations
    of a trajectory. The mask argument is used fot select these configurations.
    THIS FUNCTION TAKES INTO ACCOUNT ALL POSSIBLE CONTACTS, EVEN THE NON NATIVE ONES

    Therefore, it takes as argument the distance threshold below which two particles are
    considered in contact (instead of the modiied contact map; this is the reason behind the
    redefinition of the formed_contacts function)
    '''

    # TODO: attenzione qua: questa definizione per le distanze considera anche
    # residui molto vicini lungo la catena. Questo non ha effetto nel GEmax non nativo
    # ma lo ha nel GE medio non nativo!!
    def formed_contacts(thr, ca_pos, gamma):
        n = ca_pos.shape[0]
        # Compute distanze matrix
        self_distances = distances.self_distance_array(ca_pos)
        sq_dist_arr = np.zeros((n, n))
        triu = np.triu_indices_from(sq_dist_arr, k=1)
        sq_dist_arr[triu] = self_distances

        return (sq_dist_arr <= gamma*thr)

    GE_traj = []
    for i in range(u.trajectory.n_frames):
        if mask[i]:
            print(f'Computing GE for config: {i}')
            ca_pos = u.trajectory[i].positions
            cm_config = formed_contacts(thr, ca_pos, gamma)
            GE_traj.append(GE_loops(cm_config, ca_pos, m0))

    return GE_traj

