'''Collection of function for the computation of
native contacts formed during a simulation
'''
from typing import Tuple

import numpy as np
import MDAnalysis as mda
from MDAnalysis.analysis import distances

def formed_contacts(
    cm_native: np.ndarray, 
    ca_pos: np.ndarray, 
    gamma: float
    ) -> np.ndarray:
    '''Returns a contact map where each entry is True or False whenever a contact is formed or not

    Args:
        cm_native (2D numpy array): contact map of the native state of the protein
        ca_pos (list of list): list of the 3D coordinates of the carbons alpha
        gamma (float): parameter to define contact

    Returns:
        2D numpy array
    '''
    # load modified version of the native contact map
    cm_native_mod = np.triu(cm_native)
    cm_native_mod[cm_native_mod == 0] = -1
    n = cm_native.shape[0]

    # Compute distanze matrix
    self_distances = distances.self_distance_array(ca_pos)
    sq_dist_arr = np.zeros((n, n))
    triu = np.triu_indices_from(sq_dist_arr, k=1)
    sq_dist_arr[triu] = self_distances
    
    return (sq_dist_arr <= gamma*cm_native_mod)

def formed_contacts_timeseries(
    cm_native: np.ndarray,
    u: mda.Universe,
    gamma: float,
    begin: int,
    end: int
    ) -> np.ndarray:
    '''Returns the series of matrix representing formed contacts as a function
    of time.

    Args:
        cm_native (2D numpy array): contact map of the native state of the protein
        u (MDAnalysis Universe): Universe representing the trajectory
        gamma (float): parameter to define contact
        begin, end (int): start and end of the trajectory range to analyse
        If end == 0, the whole trajectory is considered

    Returns:
        1D Numpy array: the contact timeseries
    '''

    if end > u.trajectory.n_frames or end == 0:
        end = u.trajectory.n_frames

    n_res = cm_native.shape[0]
    timeseries = np.zeros((end-begin, n_res, n_res))

    for i in range(begin, end):
        ca_pos = u.trajectory[i].positions
        timeseries[i] = formed_contacts(cm_native, ca_pos, gamma)

    return timeseries

def reduced_Q(
    u:mda.Universe, 
    cm_a: np.ndarray, 
    cm_b: np.ndarray, 
    gamma: float
    ) -> Tuple[np.ndarray, np.ndarray]:
    '''Returns the reduced Q reaction coordinates by separating the
    set of native contacts into two groups defined by cm_a and cm_b
    
    Please note that this function do not assume that the two set a and b
    of native contacts are one the complementary of the other.
    Nevertheless cm's are assumed to be upper triangular matrices.

    Args:
        u (MDAnalysis.Universe): universe of the trajectory to analyse
        cm_a (2D Numpy array): native contact map with only contacts belonging
        to the a group
        cm_b (2D Numpy array): native contact map with only contacts belonging
        to the b group
        gamma (float): to define a contact

    Return:
        Qa, Qb (1D Numpy array each): timeseries of these reduced Q
    '''

    # compute number of contacts per group
    total_contacts_a = (cm_a != 0).sum()
    total_contacts_b = (cm_b != 0).sum()

    # Modify cm
    cm_mod_a = cm_a.copy()
    cm_mod_a[cm_a == 0] = -1 
    cm_mod_b = cm_b.copy()
    cm_mod_b[cm_b == 0] = -1 

    n = cm_mod_a.shape[0] # dim contact matrix

    # computing Qs
    Qa = []
    Qb = []
    for i in range(u.trajectory.n_frames):

        pos_ts = u.trajectory[i].positions
        # flattened upper triangular part of the distance matrix
        self_distances = distances.self_distance_array(pos_ts)

        sq_dist_arr = np.zeros((n, n))
        triu = np.triu_indices_from(sq_dist_arr, k=1)
        sq_dist_arr[triu] = self_distances

        Qa_num = (sq_dist_arr <= gamma*cm_mod_a)
        Qb_num = (sq_dist_arr <= gamma*cm_mod_b)
        Qa.append(Qa_num.sum()/total_contacts_a)
        Qb.append(Qb_num.sum()/total_contacts_b)

    return (np.array(Qa), np.array(Qb))